import logging,asyncio
# Suppress logs below WARNING level
logging.basicConfig(level=logging.WARNING)
from abstract_apis import make_request,asyncPostRequest
from ..utils import *

async def SolcatcherApiCall(url=None,endpoint=None, *args, **kwargs):
    """
    Asynchronously resolves arguments and makes a POST request to Solcatcher Flask URL.
    """
    # Resolve coroutines in args
    resolved_args = [await arg if asyncio.iscoroutine(arg) else arg for arg in args]
    
    # Resolve coroutines in kwargs
    resolved_kwargs = {
        k: await v if asyncio.iscoroutine(v) else v
        for k, v in kwargs.items()
    }

    # Make the asynchronous POST request
    return await asyncPostRequest(
        url=getSolcatcherFlaskUrl(),
        endpoint=endpoint,
        data={"args": resolved_args, **resolved_kwargs}
    )

def makeSolcatcherRequest(url, data=None, headers=None, get_post=None, endpoint=None, status_code=False, raw_response=False, response_result=None, load_nested_json=True, auth=None, *args, **kwargs):
    """
    Generic function to make a request to a Solcatcher endpoint.
    """
    request_data = {"args": args, **kwargs}  # Merge args and kwargs into the data payload
    return make_request(
        url=url,
        data=data or request_data,  # Use provided data or the constructed request_data
        headers=headers,
        get_post=get_post,
        endpoint=endpoint,
        status_code=status_code,
        raw_response=raw_response,
        response_result=response_result,
        load_nested_json=load_nested_json,
        auth=auth
    )

def SolcatcherRequest(endpoint,get_post=None, *args, **kwargs):
    """
    Makes a POST request to a Typescript endpoint.
    """
    get_post = get_post or 'POST'
    url = getSolcatcherUrl()
    return makeSolcatcherRequest(url=url, endpoint=endpoint, get_post=get_post , *args, **kwargs)


def FlaskRequest(endpoint,get_post=None, *args, **kwargs):
    """
    Makes a POST request to a Typescript endpoint.
    """
    get_post = get_post or 'POST'
    url = getSolcatcherFlaskUrl()
    return makeSolcatcherRequest(url=url, endpoint=endpoint,get_post=get_post , *args, **kwargs)

def TypescriptRequest(endpoint,get_post=None, *args, **kwargs):
    """
    Makes a POST request to a Typescript endpoint.
    """
    get_post = get_post or 'POST'
    url = getSolcatcherTsUrl()
    return makeSolcatcherRequest(url=url, endpoint=endpoint,get_post=get_post , *args, **kwargs)




import struct,base58,base64
from .logManager import *
def parse_data(program_data,all_js={}):
    # Step 1: Base64 decode the input program data
    decoded_data = base64.b64decode(program_data)
    
    offset = 8  # Adjusted offset to skip the first 8 bytes
    parsed = all_js
    
    try:
        # 32-byte publicKey (mint)
        mint = decoded_data[offset:offset + 32]
        parsed['mint'] = base58.b58encode(mint).decode('utf-8')
        offset += 32
        
        # 8-byte u64 (solAmount)
        parsed['solAmount'], = struct.unpack_from('<Q', decoded_data, offset)
        offset += 8
        
        # 8-byte u64 (tokenAmount)
        parsed['tokenAmount'], = struct.unpack_from('<Q', decoded_data, offset)
        offset += 8
        
        # 1-byte bool (isBuy)
        parsed['isBuy'], = struct.unpack_from('<?', decoded_data, offset)
        offset += 1
        
        # 32-byte publicKey (user)
        user = decoded_data[offset:offset + 32]
        parsed['user_address'] = base58.b58encode(user).decode('utf-8')
        offset += 32
        
        # 8-byte i64 (timestamp)
        parsed['timestamp'], = struct.unpack_from('<q', decoded_data, offset)
        offset += 8
        
        # 8-byte u64 (virtualSolReserves)
        parsed['virtualSolReserves'], = struct.unpack_from('<Q', decoded_data, offset)
        offset += 8
        
        # 8-byte u64 (virtualTokenReserves)
        parsed['virtualTokenReserves'], = struct.unpack_from('<Q', decoded_data, offset)
        
    except struct.error as e:
        print(f"Struct error: {e}")
        return None
    return parsed
def get_program_data(response, all_js={}):
    """
    Extract and parse program data from logs.
    """
    try:
        logs = get_logs(response)
        for log in logs:
            if log.startswith('Program data: '):
                base64_data = log[len('Program data: '):]
                all_js = parse_data(base64_data, all_js)
    except Exception as e:
        logger.error(f"Error parsing program data: {e}")
    return all_js



from abstract_utilities import SingletonMeta
from abstract_solana import make_call,get_rpc_dict,Client,abstract_solana_rate_limited_call,RateLimiter,get_insert_list
from .solcatcherCalls import *
from ..utils import *
import requests,json
from abstract_apis import postRpcRequest,get_async_response,get_json_response,make_request,get_url,make_endpoint,get_headers,get_async_response,get_headers
from abstract_database import get_from_db,get_table,get_db_vars,connectionManager,get_first_row_as_dict
import json,asyncio
def call_rate_limit(method=None,status_code=False):
  return FlaskRequest(endpoint='rate_limit',method=get_method(method),get_post='POST',response_result='url',status_code=status_code)
def call_log_response(method=None,response=None,retry_after=None):
  return requests.post(url='https://solcatcher.io/flask/log_response',data=json.dumps({"method":get_method(method),"response":get_resp(response),"retry_after":retry_after}),headers=get_headers())
class CheckSolcatcher(metaclass=SingletonMeta):
  def __init__(self,env_path=None, dbType=None, dbName=None):
    if not hasattr(self, 'initialized'):
      self.rateLimiter = RateLimiter()
      self.initialized = True
      self.solcatcher_on = True
      self.solcatcher_db_on=False
      url,self.status_code = call_rate_limit('test_method',True)
      self.env_path= env_path or get_env_path()
      if self.status_code != 200 and url == None:
        self.solcatcher_on=False
   
      self.dbType = dbType or 'database'
      self.dbName = dbName or 'partners'
      self.tables=get_insert_list()
      #input(get_first_row_as_dict(tableName=self.tables[5].get('tableName'), rowNum=1))
      self.dbVars = get_db_vars(env_path=self.env_path, dbType=self.dbType, dbName=self.dbName)
      protocol = 'postgres'
      if 'rabbit' in self.dbType.lower():
        protocol = 'amqp'
      self.dbVars['dburl'] = f"{protocol}://{self.dbVars['user']}:{self.dbVars['password']}@{self.dbVars['host']}:{self.dbVars['port']}/{self.dbVars['dbname']}"
      #self.conn_mgr = connectionManager(env_path=self.env_path, dbType=self.dbType, dbName=self.dbName, tables=self.tables, dbVars=self.dbVars)
      #input('asdffawdssdf')
      #conn = self.conn_mgr.check_conn(env_path=self.env_path, dbType=self.dbType, dbName=self.dbName)

  def get_rate_url(self,method=None,status_code=False):
    method = get_method(method)
    if self.solcatcher_on:
      try:
        url,status_code =call_rate_limit(method,True)
        if status_code == 200:
          return url
      except:
        pass
    return CheckSolcatcher().rateLimiter.get_url(method)
  def log_response(self,method=None,response=None,status_code=False):
    if self.solcatcher_on:
      try:
        return call_log_response(method=method,response=response)
      except:
        pass
    return CheckSolcatcher().rateLimiter.log_response(method=method,response=response)
  def connect_db(self,env_path=None, dbType=None, dbName=None):
     try:
       conn = create_connection(env_path=env_path, dbType=dbType, dbName=dbName)
       return conn
     except:
        pass
  def call_solcatcher_api(self,method,*args,**kwargs):
    if self.solcatcher_db_on:
      try:
        return call_solcatcher_db_api(method,*args,**kwargs)
      except:
        pass
    return abstract_solana_rate_limited_call(method,*args,**kwargs)
##    return abstract_solana_rate_limited_call(method,*args,**kwargs)
def get_method(method=None):
  return method or 'default_method'
class clientMgr(metaclass=SingletonMeta):
    def __init__(self):
        if not hasattr(self, 'initialized'):  # Prevent reinitialization
            self.initialized = True
            self.client = Client()
            self.functions = {}

    def get_client_function(self, method):
        if method not in self.functions:
            self.functions[method] = getattr(self.client, method, None)
        return self.functions.get(method)
    def inspect_function(self, method_name):
        try:
            if hasattr(self.client, method_name):
                func = self.get_client_function(method)
               
                
                if func and callable(func):
                    try:
                        # Retrieve the signature without calling the method
                        signature = inspect.getargspec(func).args
                        parameters = list(signature.parameters.keys())

                        
                        return parameters
                    except ValueError:
                        return f"Could not retrieve signature for method {method_name}."
                else:
                    return f"{method_name} not found or not callable."
            else:
                return f"{method_name} not found in client."
        except:
            return f"couldnt get the signature for {method_name}"
    def get_body(self,method,*args,**kwargs):
        function = self.get_client_function(method)
        return function(*args,**kwargs)
    def get_partial_call(self,method,body):
        url = rate_limit(method)
        response = postRpcRequest(url=url,**body)
        log_response(method, response)
        return response
    def call_solana(self,method,*args,**kwargs):
        body = self.get_body(method,*args,**kwargs)
        response = self.get_partial_call(method,body)
        return response
def call_solcatcher_api(method,*args,**kwargs):
  return CheckSolcatcher().call_solcatcher_api(method,*args,**kwargs)
def call_solana(method,*args,**kwargs):
  return clientMgr().call_solana(method,*args,**kwargs)
def get_body(method,*args,**kwargs):
  return clientMgr().get_body(method,*args,**kwargs)
def partial_call_solana(method,body):
  return clientMgr().get_partial_call(method,body)
def fetch_from_db(*args,**kwargs):
    return get_async_response(getSolcatcherPost,'fetch_from_db',*args,**kwargs)
def insert_db(*args,**kwargs):
    return get_async_response(getSolcatcherPost,'insert_into_db',*args,**kwargs)
def get_db_params(method,*args,**kwargs):
    dbName=kwargs.get('dbName','solcatcher')
    insertTable = get_table(method,get_insert_list())
    if isinstance(insertTable,dict):
        freshCall = insertTable.get('freshCall')
        tableName=insertTable.get('tableName')
        insertName=insertTable.get('insertName')
        rpcCall = insertTable.get('rpcCall')
        return dbName,insertTable,freshCall,tableName,insertName,rpcCall
def call_solcatcher_db_api(method,*args,**kwargs):
    dbName,insertTable,freshCall,tableName,insertName,rpcCall = get_db_params(method,*args,**kwargs)
    if rpcCall:
        rpc_dict = get_body(method,*args,**kwargs)
        searchValue= rpc_dict.get('params')[0]
        insertValue  = get_from_db(freshCall,tableName,searchValue,fetch_from_db)
        if insertValue is None:
            insertValue = partial_call_solana(method,rpc_dict)
            rate_limit(method=method,response = {"result":insertValue})
            
            if isinstance(insertValue,dict) and insertValue.get('error'):
                return insertValue
            if isinstance(searchValue,dict):
                searchValue = generate_data_hash(insertName,insertValue)
            insert_db(dbName,insertTable,key=searchValue, value=insertValue,tableName=tableName)
    else:
        searchValue= args[0] if args else None
        insertValue  = get_from_db(freshCall,tableName,searchValue,fetch_from_db)
        if insertValue is None:
            insertValue = rate_limit_solcatcher_api(method,*args,**kwargs)
            rate_limit(method,insertValue)
            insert_db(dbName,insertTable,key=searchValue, value=insertValue,tableName=tableName)
    return insertValue
def convert_to_lower(string_obj):
    return ''.join(f"_{char.lower()}" if char.isupper() else char for char in str(string_obj))

def abstract_solcatcher_rate_limited_call(method, *args, **kwargs):
    # Build the request body
    if '-' not in method:
      method = convert_to_lower(method)
    body = get_rpc_dict(method, *args, **kwargs)
    body_method = body.get('method')
    url = call_rate_limit(method=body_method)
    response,status_code,retry_after = make_call(url,body)   
    call_log_response(method=body_method, response=response,retry_after=retry_after)
    if "api.mainnet-beta.solana.com" in url and status_code == 429:
        url = call_rate_limit(method='get_url_2')
        response,status_code,retry_after = make_call(url,body)
        call_log_response(method=body_method, response=response)
    return response



from utilities import getEndpointUrl,try_json_dumps,getSolcatcherTsUrl,getSolcatcherFlaskUrl,getSolcatcherUrl,get_method,get_url
from abstract_utilities import eatAll,make_list
from abstract_apis import make_request,get_headers,postRequest
import requests
def getEndpointUrl(endpoint=None,url=None):
  url = url or getSolcatcherFlaskUrl()
  url = eatAll(url,['/'])
  endpoint = eatAll(endpoint,['/'])
  return f"{url}/{endpoint}"
def get_datas(data=None,*args,**kwargs):
  data = data or kwargs
  if not isinstance(data,dict):
    data={"args":args}
  data = try_json_dumps(data)
  return data
def makeSolcatcherRequest(url,endpoint,get_post='POST',*args,**kwargs):
  return make_request(url=url,endpoint=endpoint,get_post=get_post,*args,**kwargs)
def postSolcatcherRequest(url,endpoint,*args,**kwargs):
  url=getSolcatcherUrl()
  return makeSolcatcherRequest(url=url,endpoint=endpoint,get_post='POST',*args,**kwargs)
def getSolcatcherRequest(endpoint,*args,**kwargs):
  url=getSolcatcherUrl()
  return makeSolcatcherRequest(url=url,endpoint=endpoint,get_post='GET',*args,**kwargs)
def postFlaskRequest(endpoint,*args,**kwargs):
  url=getSolcatcherFlaskUrl()
  return postSolcatcherRequest(url=url,endpoint=endpoint,get_post='POST',*args,**kwargs)
def getFlaskRequest(endpoint,*args,**kwargs):
  url=getSolcatcherFlaskUrl()
  return getSolcatcherRequest(url=url,endpoint=endpoint,get_post='GET',*args,**kwargs)
def postTypescriptRequest(endpoint,get_post='POST',*args,**kwargs):
  url=getSolcatcherFlaskUrl()
  return postSolcatcherRequest(url=url,endpoint=endpoint,*args,**kwargs)
def getTypescriptRequest(endpoint,get_post='GET',*args,**kwargs):
  url=getSolcatcherFlaskUrl()
  return getSolcatcherRequest(url=url,endpoint=endpoint,*args,**kwargs)

response = postFlaskRequest(endpoint='rate_limit',data={"method":get_method()},response_result='url')
input(response)
response = getFlaskRequest(endpoint='log_response',data={"method":get_method(),'response_data':response})
input(response)



from psycopg2.extras import Json,execute_values
from datetime import datetime, timezone
from .logManager import *
from .db_credentials import DB_CONFIG
def handle_wallet_signatures(creator_wallet):
    """
    Handles checking, inserting, or updating wallet and signature data.
    """
    signatures = get_signatures(creator_wallet)

    # Convert signatures into a list of strings (PostgreSQL-compatible array format)
    signature_list = [entry['signature'] for entry in signatures]

    # Query to check if wallet address exists
    check_query = """
    SELECT id, signature, signatures FROM transaction_logs
    WHERE user_address = %s;
    """
    
    # Insert/Update queries
    insert_query = """
    INSERT INTO transaction_logs (
        user_address, signature, signatures, timestamp
    )
    VALUES (%s, %s, %s, %s)
    RETURNING *;
    """
    
    update_query = """
    UPDATE transaction_logs
    SET signatures = COALESCE(signatures, %s)
    WHERE id = %s
    RETURNING *;
    """

    results = []

    try:
        with psycopg2.connect(**DB_CONFIG) as conn:
            with conn.cursor() as cur:
                # Check if the wallet address exists in the database
                cur.execute(check_query, (creator_wallet,))
                rows = cur.fetchall()

                if rows:
                    # If the wallet exists, check and update signatures
                    for row in rows:
                        row_id, existing_signature, existing_signatures = row
                        if not existing_signatures:
                            # Update if signatures are missing
                            cur.execute(update_query, (signature_list, row_id))
                            updated_row = cur.fetchone()
                            results.append(updated_row)
                        else:
                            # Signatures already exist
                            results.append(row)
                else:
                    # Wallet address not found; insert a new row
                    cur.execute(
                        insert_query,
                        (creator_wallet, None, signature_list, datetime.now(timezone.utc)),
                    )
                    new_row = cur.fetchone()
                    results.append(new_row)

                conn.commit()

    except Exception as e:
        print(f"Error handling wallet signatures: {e}")

    return results
def insert_transaction_log(data):
    """
    Inserts or updates a transaction log in the database.
    """
    # Provide default values for any missing keys
    default_data = {
        "signature": None,
        "mint": None,
        "sol_amount": None,
        "token_amount": None,
        "is_buy": None,
        "user_address": None,
        "timestamp": datetime.now(timezone.utc),  # Use current time with timezone awareness
        "virtual_sol_reserves": None,
        "virtual_token_reserves": None,
        "metadata": None,
        "logNotification": None,
        "signatures": None,
        "transaction": None
    }
    # Merge with provided data
    default_data.update(data)

    # Convert `timestamp` to datetime if it's an integer (Unix timestamp)
    if isinstance(default_data["timestamp"], int):
        default_data["timestamp"] = datetime.fromtimestamp(default_data["timestamp"], tz=timezone.utc)

    query = """
    INSERT INTO transaction_logs (
        signature, mint, sol_amount, token_amount, is_buy, user_address, 
        timestamp, virtual_sol_reserves, virtual_token_reserves, metadata, 
        logNotification, signatures, transaction
    )
    VALUES (
        %(signature)s, %(mint)s, %(sol_amount)s, %(token_amount)s, %(is_buy)s, %(user_address)s, 
        %(timestamp)s, %(virtual_sol_reserves)s, %(virtual_token_reserves)s, %(metadata)s, 
        %(logNotification)s, %(signatures)s, %(transaction)s
    )
    ON CONFLICT (signature) DO UPDATE SET
        mint = COALESCE(EXCLUDED.mint, transaction_logs.mint),
        sol_amount = COALESCE(EXCLUDED.sol_amount, transaction_logs.sol_amount),
        token_amount = COALESCE(EXCLUDED.token_amount, transaction_logs.token_amount),
        is_buy = COALESCE(EXCLUDED.is_buy, transaction_logs.is_buy),
        user_address = COALESCE(EXCLUDED.user_address, transaction_logs.user_address),
        timestamp = COALESCE(EXCLUDED.timestamp, transaction_logs.timestamp),
        virtual_sol_reserves = COALESCE(EXCLUDED.virtual_sol_reserves, transaction_logs.virtual_sol_reserves),
        virtual_token_reserves = COALESCE(EXCLUDED.virtual_token_reserves, transaction_logs.virtual_token_reserves),
        metadata = COALESCE(EXCLUDED.metadata, transaction_logs.metadata),
        logNotification = COALESCE(EXCLUDED.logNotification, transaction_logs.logNotification),
        signatures = COALESCE(EXCLUDED.signatures, transaction_logs.signatures),
        transaction = COALESCE(EXCLUDED.transaction, transaction_logs.transaction);
    """
    try:
        with psycopg2.connect(**DB_CONFIG) as conn:
            with conn.cursor() as cur:
                cur.execute(query, {key: Json(value) if key in ['metadata', 'logNotification', 'transaction'] else value
                                    for key, value in default_data.items()})
                conn.commit()
                print(f"Transaction log inserted/updated successfully. for {list(data.keys())}")
    except Exception as e:
        print(f"Error inserting transaction log: {e}")

def fetch_transaction_logs(limit=10, offset=0, filters=None):
    """
    Fetches transaction logs from the database with optional filters and offset.
    
    :param limit: The maximum number of rows to fetch.
    :param offset: The starting point for fetching rows.
    :param filters: A dictionary of column-value pairs to filter the logs.
    :return: A list of dictionaries containing the transaction log details.
    """
    query = """
    SELECT signature, mint, sol_amount, token_amount, is_buy, user_address, 
           timestamp, virtual_sol_reserves, virtual_token_reserves, metadata, 
           logNotification, signatures, transaction
    FROM transaction_logs
    """
    
    where_clause = []
    params = []

    # Add filters dynamically
    if filters:
        for column, value in filters.items():
            where_clause.append(f"{column} = %s")
            params.append(value)

    # Add WHERE clause if filters exist
    if where_clause:
        query += " WHERE " + " AND ".join(where_clause)
    
    # Add ORDER BY, LIMIT, and OFFSET
    query += " ORDER BY timestamp DESC LIMIT %s OFFSET %s;"
    params.extend([limit, offset])

    try:
        with psycopg2.connect(**DB_CONFIG) as conn:
            with conn.cursor() as cur:
                cur.execute(query, tuple(params))
                results = cur.fetchall()
                colnames = [desc[0] for desc in cur.description]
                return [dict(zip(colnames, row)) for row in results]
    except Exception as e:
        print(f"Error fetching transaction logs: {e}")
        return []
def insert_into_table(query,data=None):
    try:
        # Connect to the database
        with psycopg2.connect(**DB_CONFIG) as conn:
            with conn.cursor() as cur:
                # Execute the query
                if data is not None:
                    cur.execute(query, data)
                else:
                    cur.execute(query)
                conn.commit()
                print(f"Transaction log inserted/updated successfully for {list(data.keys())}")
    except Exception as e:
        print(f"Error inserting transaction log: {e}")




from .utils import get_socket_url,get_websocket_params
from .programData import *
from .db_credentials import DB_CONFIG,RABBIT_CONFIG
import pika,logging,threading,websockets,asyncio
from abstract_solcatcher import TypescriptRequest
logger = logging.getLogger(__name__)
def process_websocket_message(message):
#    """
#    Process an individual WebSocket message.
#    """
#    try:
# Confirm and retrieve log by signature
    logNotification = log_mgr.confirm_log(logNotification=message)
    if not logNotification:
        raise ValueError("Invalid or empty logNotification")

    signature = get_signature_from_log(log=logNotification)
    if not signature:
        raise ValueError("Missing signature in logNotification")

    # Parse program data
    all_js = get_program_data(logNotification, {"signature": signature})
    if not all_js:
        raise ValueError("Failed to parse program data from logNotification")
    mintAddress = all_js.get('mint')
    if mintAddress:
        all_js['metaData'] = TypescriptRequest(endpoint='getMetaData', mintAddress=mintAddress)
    
    # Check for 'user_address' in all_js
    user_address = all_js.get('user_address')
    if not user_address:
        raise ValueError("'user_address' not found in parsed data")
    
    # Insert parsed data and handle wallet signatures
    insert_transaction_log(all_js)
    handle_wallet_signatures(user_address)

    # Fetch from the database for verification
    #fetched = fetch_transaction_logs(filters={'user_address': user_address})
    #print(f"Fetched logs for user_address {user_address}")



#    except Exception as e:
#        logger.error(f"Error processing WebSocket message: {e}")
    
# RabbitMQ Consumer
def consume_rabbitmq():
    try:
        connection_params = pika.ConnectionParameters(
            host=RABBIT_CONFIG.get('host'),
            credentials=pika.PlainCredentials(RABBIT_CONFIG.get('user'), RABBIT_CONFIG.get('password'))
        )
        connection = pika.BlockingConnection(connection_params)
        channel = connection.channel()
        channel.queue_declare(queue=RABBIT_CONFIG.get('queue'), durable=True)

        logger.info(f"Connected to RabbitMQ. Listening on queue: {RABBIT_CONFIG.get('queue')}")

        def callback(ch, method, properties, body):
            logger.info(f"Received RabbitMQ message: {body.decode()}")
            try:
                message_data = json.loads(body)
                logger.info(f"Parsed RabbitMQ message: {message_data}")
            except json.JSONDecodeError as e:
                logger.error(f"Failed to parse RabbitMQ message: {e}")

            ch.basic_ack(delivery_tag=method.delivery_tag)
        
        channel.basic_consume(queue=RABBIT_CONFIG.get('queue'), on_message_callback=callback)
        channel.start_consuming()

    except Exception as e:
        logger.error(f"Error in RabbitMQ consumer: {e}")
# WebSocket Listener
async def connect_to_websocket():
    """
    Connect to the WebSocket and process incoming messages.
    """
    
    
    while True:
        #try:
            async with websockets.connect(get_socket_url()) as websocket:
                await websocket.send(get_websocket_params())
                response = await websocket.recv()
                logger.info(f"Subscribed to logs: {response}")

                async for message in websocket:
                    if get_log_value_from_key(message,'err') == None:
                        process_websocket_message(message)

        #except Exception as e:
        #    logger.error(f"WebSocket error: {e}")
        #    await asyncio.sleep(5)  # Retry after a short delay
def run_solcatcher():
    rabbitmq_thread = threading.Thread(target=consume_rabbitmq, daemon=True)
    rabbitmq_thread.start()
    asyncio.run(connect_to_websocket())



import struct,base58,base64,time,json
from abstract_utilities import safe_json_loads,get_any_value,SingletonMeta,make_list
from ...database_calls import abstract_solcatcher_rate_limited_call, TypescriptRequest,call_rate_limit,call_log_response,FlaskRequest,get_body
from abstract_apis import postRpcRequest
def get_pumpfun_program_wallet():
    return "6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"
def get_wallet(wallet_type=None):
    wallet_types = {"pumpFun":get_pumpfun_program_wallet()}
    wallet_type = wallet_type or "pumpFun"
    for walletType,wallet_address in wallet_types.items():
        if walletType.lower() == wallet_type.lower():
            return wallet_address
    return list(wallet_types.values())[0]
def get_socket_url():
    return "wss://api.mainnet-beta.solana.com/"
def get_websocket_params(jsonrpc=None,id=None,method=None,params=None,mentions=None,commitment=None,wallet_type=None,wallet_address=None):
    jsonrpc = str(float(jsonrpc or 2))
    id = int(id or 1)
    method = method or "logsSubscribe"
    wallet_type=wallet_type or "pumpFun"
    wallet_address = get_wallet(wallet_type)
    mentions = make_list(mentions or [wallet_address])
    commitment = commitment or "processed"
    params = make_list(params or [{"mentions": mentions},{"commitment": commitment}])
    return json.dumps({
        "jsonrpc": jsonrpc,
        "id": id,
        "method": method,
        "params": params})
def get_signatures(creator_wallet):
    signatures = abstract_solcatcher_rate_limited_call('get_signatures_for_address', creator_wallet, limit=1000)
    return signatures
def test_it():
    
    logd = {'jsonrpc': '2.0', 'method': 'logsNotification', 'params': {'result': {'context': {'slot': 302009204}, 'value': {'signature': 'poSi4pY6MTmK79W4FyqSUtpGSfgoPzu77mwVYkbktqp4wyq4qtLw8qpHjCAnUUb8Q9i8nJGTEYyJxUp2hwMiJtu', 'err': 'sdffsdf', 'logs': ['Program ComputeBudget111111111111111111111111111111 invoke [1]', 'Program ComputeBudget111111111111111111111111111111 success', 'Program ComputeBudget111111111111111111111111111111 invoke [1]', 'Program ComputeBudget111111111111111111111111111111 success', 'Program BSfD6SHZigAfDWSjzD5Q41jw8LmKwtmjskPH9XW1mrRW invoke [1]', 'Program log: Instruction: PumpSell', 'Program 6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P invoke [2]', 'Program log: Instruction: Sell', 'Program TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA invoke [3]', 'Program log: Instruction: Transfer', 'Program TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA consumed 4645 of 107609 compute units', 'Program TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA success', 'Program 6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P invoke [3]', 'Program 6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P consumed 2003 of 99479 compute units', 'Program 6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P success', 'Program data: vdt/007mYe7oh9/peHrXFNq4S++LHVOYSLzIMMrBsxC18cwELmrCxnbY0H4AAAAAQLJ375tAAAAAqudgL7+RixOJeYKxNmsEFpIy7yATPAuxu60l99nvOzpBWzpnAAAAAD6IZfwGAAAAusCYV7/PAwA+3EEAAAAAALoohgsu0QIA', 'Program 6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P consumed 36417 of 132167 compute units', 'Program 6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P success', 'Program 11111111111111111111111111111111 invoke [2]', 'Program 11111111111111111111111111111111 success', 'Program BSfD6SHZigAfDWSjzD5Q41jw8LmKwtmjskPH9XW1mrRW consumed 57186 of 149700 compute units', 'Program BSfD6SHZigAfDWSjzD5Q41jw8LmKwtmjskPH9XW1mrRW success', 'Program 11111111111111111111111111111111 invoke [1]', 'Program 11111111111111111111111111111111 success']}}, 'subscription': 55791892}}
    start = time.time()
    error = get_log.get_err(log=logd)
    long = time.time() - start
    input(f"{long} seconds seconds to get {error}")
    start = time.time()
    error = get_any_value(logd,'err')
    long = time.time() - start
    input(f"{long} seconds seconds to get {error}")
    log = logd
    start = time.time()
    for each in ['params','result','value','err']:
        log = log.get(each)
    new = time.time() - start
    input(f"{new} seconds to get {log}")
    log = logd
    start = time.time()
    log = log.get('params').get('result').get('value').get('err')
    new = time.time() - start
    input(f"{new} seconds to get {log}")
    print(f"long was {(long/new)*100}% slower")
    input(log)



import asyncio,requests,json
from abstract_utilities import *
from abstract_solana import RateLimiter
from abstract_apis import make_request,get_url,make_endpoint,get_headers,get_async_response
def try_solcatcher_io(method = None):
  method = get_method(method=method)
  try:
    url,status_code = postFlaskRequest(url=getSolcatcherFlaskUrl(),endpoint='get_rate_limit',data=json.dumps({"method":method}),status_code=True)
    return url,status_code 
  except:
    pass
  return None,500
def try_log_response(method = None,response=None):
  method = get_method(method=method)
  response = get_resp(response=response)
  try:
    response = postFlaskRequest(url=getSolcatcherFlaskUrl(),endpoint='log_response',data=json.dumps({"method":method,"response":response}),status_code=True)
    return response 
  except:
    pass
class CheckSolcatcher(metaclass=SingletonMeta):
  def __init__(self):
    if not hasattr(self, 'initialized'):
      self.initialized = True
      self.solcatcher_on = True
      url,self.status_code = try_solcatcher_io()
      if self.status_code != 200 and get_url(url=url) == None:
        self.solcatcher_on=False
      self.rateLimiter = RateLimiter()
  def get_rate_url(self,method=None):
    if self.solcatcher_on:
      url,response_code = try_solcatcher_io(method = method)
    if self.solcatcher_on == False or self.status_code != 200:
       self.solcatcher_on = False
       method = get_method(method=method)
       url = self.rateLimiter.get_url(method=method)
    return get_url(url=url)
  def log_response(self,method=None,response=None):
    if self.solcatcher_on:
       try_log_response(method = method,response=response)
    if self.solcatcher_on == False:
      method = get_method(method=method)
      response = get_resp(response=response)
      self.rateLimiter.log_response(method=method,response=response)  
def rate_limit(method=None,response=None):
  if response is not None:
      CheckSolcatcher().log_response(method=method,response=response)
      return
  return CheckSolcatcher().get_rate_url(method=method)



from .utils import getEndpointUrl,get_async_response
from abstract_apis import getRequest,requests,asyncPostRpcRequest,asyncPostRequest,asyncGetRequest,get_headers
from abstract_utilities import make_list
import json
async def dbinsert(tableName=None,searchValue=None,insertValues=None,dbName=None,dbType=None,**kwargs):
    response = await asyncPostRequest(url=getEndpointUrl("dbInsert"), data={"tableName":tableName, "searchValue":searchValue,"insertValues":insertValues,"dbName":dbName,"dbType":dbType}, **kwargs)
    return response
def dbInsert(tableName=None,searchValue=None,insertValues=None,dbName=None,dbType=None,**kwargs):
    return get_async_response(dbinsert,tableName=tableName,searchValue=searchValue,insertValues=insertValues,dbName=dbName,dbType=dbType,**kwargs)

async def dbsearch(tableName=None,searchValue=None,dbName=None,dbType=None,**kwargs):
    response = await asyncPostRequest(url=getEndpointUrl("dbSearch"), data={"tableName":tableName, "searchValue":searchValue,"dbName":dbName,"dbType":dbType},**kwargs)
    return response

def dbSearch(tableName=None,searchValue=None,dbName=None,dbType=None,**kwargs):
    return get_async_response(dbsearch, tableName=tableName,searchValue=searchValue,dbName=dbName,dbType=dbType,**kwargs)

async def asyncMakeLimitedDbCall(*args,tableName=None,searchValue=None,function=None,dbName=None,dbType=None,**kwargs):
    tableName=tableName or method
    response = await dbsearch(tableName=tableName,searchValue=searchValue,dbName=dbName,dbType=dbType,status_code=True)
    if response and response[1] == 200 and response[0]:
      print(f'search of {method} in table {tableName} successful ')
      return response
    if function:
        insertValue = await function(*args,**kwargs)
    else:
        insertValue = await asyncMakeLimitedCall(**kwargs)
    
    response = await dbinsert(tableName=tableName,searchValue=searchValue,insertValue=insertValue,dbName=dbName,dbType=dbType,status_code=True)
    return response
def makeLimitedDbCall(*args,tableName=None,searchValue=None,function=None,dbName=None,dbType=None,**kwargs):
    return get_async_response(asyncMakeLimitedDbCall, *args,tableName=tableName,searchValue=searchValue,function=function,dbName=dbName,dbType=dbType,**kwargs)


async def makeDbCall(*args,method=None,tableName=None,searchValue=None,insertValue=None,function=None,**kwargs):
    response = await dbsearch(tableName,searchValue)
    if response:
        return response
    insertValue = await function(*args,**kwargs)
    await dbinsert(tableName,searchValue,insertValue)
    return insertValue

async def asyncMakeLimitedCall(method=None, params=[],**kwargs):
    urls = await async_get_rate_limit_url(method)
    response = await asyncPostRpcRequest(
        url=urls.get('url'), method=method, params=params, status_code=True, response_result='result'
    )
    
    if response[1] == 429:
        response = await asyncPostRpcRequest(
            url=urls.get('url2'), method=method, params=params, response_result='result', status_code=True
        )
    response = response[0]
    await async_log_response(method,response)
    return response
def makeLimitedCall(method=None, params=[]):
    return get_async_response(asyncMakeLimitedCall, method, params)

async def async_get_rate_limit_url(method='default_method'):
    return await asyncGetRequest(url=getEndpointUrl("rate_limit"),data={"method":str(method)})

def get_rate_limit_url(method_name, *args, **kwargs):
    return get_async_response(async_get_rate_limit_url, method_name, *args, **kwargs)

async def async_log_response(method='default_method', response_data={}):
    return await asyncPostRequest(url=getEndpointUrl("log_response"),data={"method":str(method),"response_data":response_data})

def log_response(method_name, response_data, *args, **kwargs):
    return get_async_response(async_log_response, method_name, response_data, *args, **kwargs)



from abstract_security import *
def get_dbConfig_keys():
  return ["dbname","user","password","host","port"]
def get_rabbitmq_keys():
  return ["user","password","host","queue"]
def get_db_env_key(key,dbProgram=None,dbType=None):
  dbProgram=dbProgram or 'solcatcher'
  dbType=dbType or 'database'
  return f"{dbProgram.upper()}_{dbType.upper()}_{key.upper()}"
def get_db_env_value(key,env_path=None):
  return get_env_value(key=key,path=env_path)
def get_env_data(keys,dbProgram=None,dbType=None,env_path=None):
  dbProgram=dbProgram or 'solcatcher'
  dbType=dbType or 'database'
  db_js = {}
  for key in keys:
    db_env_key = get_db_env_key(key=key,dbProgram=dbProgram,dbType=dbType)
    db_js[key]= get_db_env_value(key=db_env_key,env_path=env_path)
  return db_js
DB_CONFIG = get_env_data(get_dbConfig_keys(),dbProgram='solcatcher',dbType='database')
RABBIT_CONFIG = get_env_data(get_rabbitmq_keys(),dbProgram='solcatcher',dbType='rabbitmq')



import json
def getSolcatcherUrl():
  return 'https://solcatcher.io'
def getSolcatcherFlaskUrl():
  return 'https://solcatcher.io/flask'
def getSolcatcherTsUrl():
  return 'https://solcatcher.io/typescript'
def getEndpointUrl(endpoint=None,url=None):
  if endpoint:
    url = eatAll(url,['/'])
    endpoint = eatAll(endpoint,['/'])
    url= f"{url}/{endpoint}"
  return url
def try_json_dumps(data):
  if isinstance(data,dict):
    try:
      data = json.dumps(data)
    except:
      pass
    return data
def get_url(url=None):
    if isinstance(url,dict):
      url = url.get('url',url)
    return url
def getCallArgs(endpoint):
  return {'getMetaData': ['signature'], 'getPoolData': ['signature'], 'getTransactionData': ['signature'], 'getPoolInfo': ['signature'], 'getMarketInfo': ['signature'], 'getKeyInfo': ['signature'], 'getLpKeys': ['signature'], 'process': ['signature']}.get(get_endpoint(endpoint))
def ifListGetSection(listObj,section=0):
  if isinstance(listObj,list):
      if len(listObj)>section:
          return listObj[section]
  return listObj
def updateData(data,**kwargs):
  data.update(kwargs)
  return data
def get_method(method=None):
  return method or 'default_method'
def get_resp(response=None):
  response = response or {}
  if isinstance(response,dict):
    response = {"response":response}
  return response
